
/*
 * cpuid.c - CPU dectection routines
 * Copyright (c) 2003, The LusitanOS Project. All rights reserved.
 *
 * maintainer(s): hybrid-2k
 * started: 2003-03-26
 * changed: 2003-06-21
 *
 * This program is free software, and you can redistribute it and/or modify it
 * under the terms of the LSD (Lusitan Source Distribution) License, which are
 * described in the following text.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files, to deal them without
 * restriction, including without limitation the rights to use, copy, modify,
 * merge, publish, distribute, sublicense, and/or sell copies of it, and to
 * permit persons to whom it is furnished to do so, subject to the following
 * conditions:
 *  - The above copyright notice and this permission notice shall be included
 *    in all copies or portions of this software.
 *  - The source code and documentation of this software shall be included
 *    along with all binary distributions and any other kind of product that
 *    uses this software or substantial portions from it.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 */

#include "kstdio.h"
#include "cpuid.h"

/******************************************************************************/
/* needed string function */
/* should we use libc string lib instead? */

int strcmp(const char *str1, const char *str2)
{
	int not_equal = 0;
	while (*str1) {
		if (!(*str2)) {
			return 1;
		} else if (*str1 == *str2) {
			str1++;
			str2++;
		} else {
			not_equal++;
			break;
		}
	}
	if (not_equal || ((unsigned char) (*str2))) {
		return -1;
	} else {
		return 0;
	}
}

/******************************************************************************/

/* Check if CPUID instruction is supported */
int hasCPUID()
{
	register int ret;
	__asm__(
		/* Get copies of EFLAGS into EAX and ECX */
		"pushf \n\t"
		"popl %%eax \n\t"
		"movl %%eax, %%ecx \n\t"

		/* Toggle the ID bit in EAX and store in EFLAGS */
		"xorl $0x200000, %%eax \n\t"
		"push %%eax \n\t"
		"popf \n\t"

		/* Get the (hopefully modified) EFLAGS */
		"pushf \n\t"
		"popl %%eax \n\t"

		/* Compare and test result */
		"xorl %%eax, %%ecx \n\t"
		"testl $0x200000, %%ecx \n\t"

		/* Return the right value */
		"xor %%eax, %%eax \n\t"
		"jnz _end \n\t"
		"incl %%eax \n\t"
		"_end: \n\t"
		:"=a"(ret)
		:
		:"%ecx"
	);
	return ret;
}

/* Display all info we can get from CPUID */
int cpuid()
{
	unsigned long VendorSign[4];
	unsigned long MaxEAX;
	char *Comp1[32];

	__asm__(
		"cpuid \n\t"
		:"=a"(MaxEAX), "=b"(VendorSign[0]), "=d"(VendorSign[1]), "=c"(VendorSign[2])
    	:"a"(0)
	);

	VendorSign[3] = 0;
	printk("\tVendor string: %s\n", VendorSign);

	/* Strings for common features */
	Comp1[0] = "FPU";	/* Floating Point Unit */
	Comp1[1] = "VME";	/* Virtual Mode Extension */
	Comp1[2] = "DE";	/* Debugging Extension */
	Comp1[3] = "PSE";	/* Page Size Extension */
	Comp1[4] = "TSC";	/* Time Stamp Counter */
	Comp1[5] = "MSR";	/* Model Specific Registers */
	Comp1[6] = "PAE";	/* Physical Address Extesnion */
	Comp1[7] = "MCE";	/* Machine Check Extension */
	Comp1[8] = "CX8";	/* CMPXCHG8 Instruction */
	Comp1[9] = "APIC";	/* On-chip APIC Hardware */
	Comp1[10] = "";		/* Reserved */
	Comp1[11] = "SEP";	/* Intel:SYSENTER/SYSEXIT or AMD:SYSCALL/SYSRET */
	Comp1[12] = "MTRR";	/* Machine Type Range Registers */
	Comp1[13] = "PGE";	/* Global Paging Extension */
	Comp1[14] = "MCA";	/* Machine Check Architecture */
	Comp1[15] = "CMOV";	/* Conditional Move Instrction */
	Comp1[16] = "PAT";	/* Page Attribute Table */
	Comp1[17] = "PSE-36";	/* 36-bit Page Size Extension */
	Comp1[23] = "MMX";	/* MMX Technology */
	Comp1[24] = "FXSR";	/* FXSAVE FXRSTOR */
	Comp1[25] = "SSE";	/* Streaming SIMD Extensions */

	int isIntel = !(strcmp((char *) VendorSign, "GenuineIntel"));
	int isAMD = !(strcmp((char *) VendorSign, "AuthenticAMD"));

	/* Features for Intel CPUs only */
	if (isIntel) {
		Comp1[18] = "PSN";	/* 96-bit Processor Serial Number */
		Comp1[19] = "CLFSH";	/* CLFLUSH Instruction */
		Comp1[20] = "";	/* Reserved */
		Comp1[21] = "DS";	/* Debug Trace Store */
		Comp1[22] = "ACPI";	/* ACPI Support */
		Comp1[26] = "SSE2";	/* Streaming SIMD Extensions 2 */
		Comp1[27] = "SS";	/* Self-Snoop */
		Comp1[28] = "HTT";	/* Hyper-Threading Technology */
		Comp1[29] = "TM";	/* Thermal Monitor Supported */
		Comp1[30] = "IA-64";	/* IA-64 capable */
		Comp1[31] = "";	/* Reserved */
	}
	/* Features for AMD CPUs only */
	else if (isAMD) {

		Comp1[18] = "";
		Comp1[19] = "MPC";	/* MultiProcessing Capable */
		Comp1[20] = "";	/* Reserved */
		Comp1[21] = "";
		Comp1[22] = "MIE";	/* AMD Multimedia Instruction Extensions */
		Comp1[26] = "";
		Comp1[27] = "";
		Comp1[28] = "";
		Comp1[29] = "X86-64";	/* Running x86-64 in Longmode */
		Comp1[30] = "3DNowExt";	/* 3DNow Instruction Extension */
		Comp1[31] = "3DNow";	/* 3DNow Instructions */
	}

	unsigned long REGEAX, REGEBX, REGECX, REGEDX;
	int dFamily, dModel, dStepping, dFamilyEx, dModelEx, dType;
	int dComp1Supported[32];
	int dBrand, dCacheLineSize, dLogicalProcessorCount, dLocalAPICID;
	unsigned long Q, C;

	if (MaxEAX >= 1) {
		__asm__("cpuid \n\t"
			:"=a"(REGEAX), "=b"(REGEBX), "=c"(REGECX), "=d"(REGEDX)
			:"a"(1)
		);

		dFamily = ((REGEAX >> 8) & 0xF);
		dModel = ((REGEAX >> 4) & 0xF);
		dStepping = (REGEAX & 0xF);
		dFamilyEx = ((REGEAX >> 20) & 0xFF);
		dModelEx = ((REGEAX >> 16) & 0xF);
		dType = ((REGEAX >> 12) & 0x7);
		for (C = 1, Q = 0; Q < 32; C *= 2, Q++) {
			dComp1Supported[Q] = (REGEDX & C) != 0 ? 1 : 0;
		}
		dBrand = REGEBX & 0xFF;
		dCacheLineSize = ((strcmp(Comp1[19], "CLFSH") == 0)
				  && (dComp1Supported[19] == 1)) ? ((REGEBX >> 8) & 0xFF) * 8 : -1;
		dLogicalProcessorCount = ((strcmp(Comp1[28], "HTT") == 0)
					  && (dComp1Supported[28] == 1)) ? ((REGEBX >> 16) & 0xFF) : -1;
		dLocalAPICID = ((REGEBX >> 24) & 0xFF);
	}

	switch (dType) {
	case 1:
		printk("\tOverDrive ");
		break;
	case 2:
		printk("\tDual ");
		break;
	default:
		printk("\t");
		break;
	}

	if (isIntel) {
		switch (dFamily) {
		case 4:
			printk("486\n");
			break;
		case 5:
			printk("Pentium\n");
			break;
		default:
			printk("Pentium%d\n", dFamily-4);
			break;
		}
	} else if (isAMD) {
		printk("CPU Type: %s\n", dType);
		switch (dFamily) {
		case 4:
			printk("AMD 486\n");
			break;
		default:
			printk("K%d\n", dFamily);
			break;
		}
	}

	printk("\tFamily %d, Model %d, Stepping %d\n", dFamily, dModel, dStepping);
	printk("\tExtended Family %d, Extended Model %d\n", dFamilyEx, dModelEx);
	printk("\tSupported Features: ");
	for (Q = 0; Q < 27; Q++) {
		if (dComp1Supported[Q])
			printk("%s ", Comp1[Q]);
	}

	printk("\n");
	/* These seem not to be working */
	/*
	printk("\tCache Line Size: %d\n", dCacheLineSize);
	printk("\tLogical Processor Count: %d\n", dLogicalProcessorCount);
	*/
	if (dFamily >= 7) printk("\tLocal APIC ID: %d\n", dLocalAPICID);

	return dFamily;
}
